/*
 *
 * Copyright  2010-2013 Balzs Tth <tbalazs@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <cuda.h>
#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>
#include <stdlib.h>
#include <iostream>


#pragma region MAP
__global__
void mapKernel(int* data, unsigned int dataSize)
{
	int id = threadIdx.x + blockIdx.x * blockDim.x;
	if(id > dataSize) return;
	int square = data[id] * data[id];
	data[id] = square;
}

void map()
{
	std::cout << "Map primitive" << std::endl;

	const size_t dataSize = 33554432; // 2^25
	int* hData = new int[dataSize];
	for(size_t i = 0; i < dataSize; ++i)
	{
		hData[i] = i;
	}

	int* dData = NULL;
	cudaMalloc(&dData, sizeof(int) * dataSize);
	cudaMemcpy(dData, hData, sizeof(int) * dataSize, cudaMemcpyHostToDevice);

	dim3 block(1024, 1, 1);
	dim3 grid((dataSize + block.x - 1) / block.x, 1, 1);
	mapKernel<<<grid, block>>>(dData, dataSize);

	cudaMemcpy(hData, dData, sizeof(int) * dataSize, cudaMemcpyDeviceToHost);

	// Evaluation
	bool passed = true;
	for(size_t i = 0; i < dataSize; ++i)
	{
		if(hData[i] != i * i)
		{
			passed = false;
			std::cerr << "Wrong value at " << i << ". Value is " << hData[i] << " but expected " << i * i << "." << std::endl;
            break;
		}
	}
	if(passed)
	{
		std::cout << "Test PASSED" << std::endl;
	} 
	else 
	{
		std::cout << "Test FAILED" << std::endl;
	}

	cudaFree(dData);
	delete[] hData;
}
#pragma endregion

#pragma region HISTOGRAM GLOBAL
__global__
void histogramGlobalKernel(int* data, int* histogram, int dataSize)
{
	int id = threadIdx.x + blockIdx.x * blockDim.x;
	if(id > dataSize) return;
	atomicAdd(&histogram[data[id]], 1);
}

void histogram_global()
{
	std::cout << "Global histogram" << std::endl;
    const size_t dataSize = 33554432; // 2^25
    const size_t histogramSize = 32;

	int *hData = new int[dataSize];
    for(size_t i = 0; i < dataSize; ++i)
    {
        hData[i] = i % histogramSize;
    }

    int *hHist = new int[histogramSize];
    memset(hHist, 0, sizeof(int)*histogramSize);

	int *dData = NULL;
	cudaMalloc(&dData, sizeof(int) * dataSize);

	int *dHist = NULL;
	cudaMalloc(&dHist, sizeof(int) * histogramSize);

	cudaMemcpy(dData, hData, sizeof(int) * dataSize, cudaMemcpyHostToDevice);
	cudaMemcpy(dHist, hHist, sizeof(int) * histogramSize, cudaMemcpyHostToDevice);

	dim3 block(1024, 1, 1);
	dim3 grid((dataSize + block.x - 1) / block.x, 1, 1);
	histogramGlobalKernel<<<grid, block>>>(dData, dHist, dataSize);

	cudaMemcpy(hHist, dHist, sizeof(int) * histogramSize, cudaMemcpyDeviceToHost);

	// reference
    int *refHist = new int[histogramSize];
    memset(refHist, 0, sizeof(int)*histogramSize);
    for(size_t i = 0; i < dataSize; ++i)
    {
        refHist[hData[i]] += 1;
    }

	bool passed = true;
    for(size_t i = 0; i < histogramSize; ++i)
    {
        if(refHist[i] != hHist[i])
        {
			passed = false;
            std::cerr << "Wrong value at " << i << ". Value is " << hHist[i] << " but expected " << refHist[i] << "." << std::endl;
            break;
        }
    }
	if(passed)
	{
		std::cout << "Test PASSED" << std::endl;
	} 
	else 
	{
		std::cout << "Test FAILED" << std::endl;
	}

	delete[] refHist;
	cudaFree(dHist);
	cudaFree(dData);
	delete[] hHist;
	delete[] hData;
}
#pragma endregion

#pragma region HISTOGRAM SHARED
__global__
void histogramLocalKernel(int* data, int* histogram, int dataSize, int histogramSize)
{
	extern __shared__ int lhistogram[];
	int id = threadIdx.x + blockIdx.x * blockDim.x;
	int lid = threadIdx.x;
	
	if(lid < histogramSize)
	{
		lhistogram[lid] = 0;
	}
	
	__syncthreads();
	
	if(id < dataSize)
	{
		atomicAdd(&lhistogram[data[id]], 1);
	}

	__syncthreads();

	if(lid < histogramSize)
	{
		atomicAdd(&histogram[lid], lhistogram[lid]);
	}
}

void histogram_local()
{
	std::cout << "Shared histogram" << std::endl;
    const size_t dataSize = 33554432; // 2^25
    const size_t histogramSize = 32;

	int *hData = new int[dataSize];
    for(size_t i = 0; i < dataSize; ++i)
    {
        hData[i] = i % histogramSize;
    }

    int *hHist = new int[histogramSize];
    memset(hHist, 0, sizeof(int)*histogramSize);

	int *dData = NULL;
	cudaMalloc(&dData, sizeof(int) * dataSize);

	int *dHist = NULL;
	cudaMalloc(&dHist, sizeof(int) * histogramSize);

	cudaMemcpy(dData, hData, sizeof(int) * dataSize, cudaMemcpyHostToDevice);
	cudaMemcpy(dHist, hHist, sizeof(int) * histogramSize, cudaMemcpyHostToDevice);

	dim3 block(1024, 1, 1);
	dim3 grid((dataSize + block.x - 1) / block.x, 1, 1);
	histogramLocalKernel<<<grid, block, histogramSize>>>(dData, dHist, dataSize, histogramSize);

	cudaMemcpy(hHist, dHist, sizeof(int) * histogramSize, cudaMemcpyDeviceToHost);

	// reference
    int *refHist = new int[histogramSize];
    memset(refHist, 0, sizeof(int)*histogramSize);
    for(size_t i = 0; i < dataSize; ++i)
    {
        refHist[hData[i]] += 1;
    }

	bool passed = true;
    for(size_t i = 0; i < histogramSize; ++i)
    {
        if(refHist[i] != hHist[i])
        {
			passed = false;
            std::cerr << "Wrong value at " << i << ". Value is " << hHist[i] << " but expected " << refHist[i] << "." << std::endl;
            break;
        }
    }
	if(passed)
	{
		std::cout << "Test PASSED" << std::endl;
	} 
	else 
	{
		std::cout << "Test FAILED" << std::endl;
	}

	delete[] refHist;
	cudaFree(dHist);
	cudaFree(dData);
	delete[] hHist;
	delete[] hData;
}
#pragma endregion

#pragma region REDUCE GLOBAL
__global__
void reduceKernel(float* data, int dataSize)
{
	int id = threadIdx.x + blockIdx.x * blockDim.x;

	for(unsigned int s = dataSize / 2; s > 0; s >>= 1)
	{
		if(id < s)
		{
			data[id] = max(data[id], data[id+s]);
		}
		__syncthreads();
	}
}

void reduce_global()
{
	std::cout << "Global reduce" << std::endl;

	const size_t dataSize = 33554432; // 2^25

	float* hData = new float[dataSize];
	for(size_t i = 0; i < dataSize; ++i)
	{
		hData[i] = rand() / (float)RAND_MAX;
	}

	float* dData = NULL;
	cudaMalloc(&dData, sizeof(float) * dataSize);
	cudaMemcpy(dData, hData, sizeof(float) * dataSize, cudaMemcpyHostToDevice);

	dim3 block(1024, 1, 1);
	dim3 grid((dataSize + block.x - 1) / block.x, 1, 1);
	reduceKernel<<<grid, block>>>(dData, dataSize);

	cudaMemcpy(hData, dData, sizeof(float) * dataSize, cudaMemcpyDeviceToHost);

	// Evaluation
    float minVal = std::numeric_limits<float>::min();
    for(size_t i = 0; i < dataSize; ++i)
    {
        minVal = std::max(minVal, hData[i]);
    }
	bool passed = true;
    if(minVal != hData[0])
    {
		passed = false;
        std::cerr << "Wrong result. Value is " << hData[0] << " but expected " << minVal << std::endl;
    }
	if(passed)
	{
		std::cout << "Test PASSED" << std::endl;
	} 
	else 
	{
		std::cout << "Test FAILED" << std::endl;
	}

	cudaFree(dData);
	delete[] hData;
}
#pragma endregion

#pragma region EXCLUSIVE SCAN
__global__
void exscanKernel(int* data, int dataSize)
{
	int id = threadIdx.x + blockIdx.x * blockDim.x;
	data[id] = (id > 0) ? data[id-1] : 0;
	__syncthreads();

	for(int s = 1; s < dataSize; s *= 2)
	{
		int tmp = data[id];
		if(id + s < dataSize)
		{
			data[id + s] += tmp;
		}
		__syncthreads();
	}

	if(id == 0)
	{
		data[0] = 0;
	}
}

void exclusive_scan()
{
	std::cout << "Global exclusive scan" << std::endl;
	const size_t dataSize = 1024;

	int* hData = new int[dataSize];
	for(size_t i = 0; i < dataSize; ++i)
	{
		hData[i] = ((float)rand() / (float)RAND_MAX) < 0.5 ? 0 : 1;
	}
	int* oData = new int[dataSize];
	memcpy(oData, hData, sizeof(int) * dataSize);

	int* dData = NULL;
	cudaMalloc(&dData, sizeof(int) * dataSize);
	cudaMemcpy(dData, hData, sizeof(int) * dataSize, cudaMemcpyHostToDevice);

	dim3 block(1024, 1, 1);
	dim3 grid((dataSize + block.x - 1) / block.x, 1, 1);
	exscanKernel<<<grid, block>>>(dData, dataSize);

	cudaMemcpy(hData, dData, sizeof(int) * dataSize, cudaMemcpyDeviceToHost);
	
	// Evaluation
    int runningSum = 0;
    for(size_t i = 0; i < dataSize; ++i)
    {
        int tmp = oData[i];
        oData[i] = runningSum;
        runningSum += tmp;
    }

	bool passed = true;
    for(size_t i = 0; i < dataSize; ++i)
    {
        if(oData[i] != hData[i])
        {
			passed = false;
            std::cerr << "Wrong value at " << i << ". Value is " << hData[i] << " but expected " << oData[i] << "." << std::endl;
            break;
        }
    }

	if(passed)
	{
		std::cout << "Test PASSED" << std::endl;
	} 
	else 
	{
		std::cout << "Test FAILED" << std::endl;
	}

}
#pragma endregion

int main()
{
	map();
	histogram_global();
	histogram_local();
	reduce_global();
	exclusive_scan();

	// Flush all GPU activity
	cudaDeviceReset();
    return 0;
}
